import matplotlib.pyplot as plt
from PIL import Image
import numpy as np
import os.path
import cv2
# from torchvision import transforms
import json
import argparse
import glob
from tqdm import tqdm

parser = argparse.ArgumentParser(add_help=False)
print(parser)

# parser.add_argument('-image_directory', default='CelebAMask-HQ/CelebA-HQ-img', type=str)
# parser.add_argument('-label_directory', default='CelebAMask-HQ/CelebAMask-HQ-mask-anno', type=str)
# parser.add_argument('-datasource_directory', default='celeb_hq_dataset-test', type=str)
parser.add_argument('-datasource_directory', default='D:\EmotionSegmentationData\celeb_hq_dataset_nose', type=str)
# parser.add_argument('-image_directory', default='D:\EmotionSegmentationData\CelebAMask-HQ\CelebA-HQ-img', type=str)
# parser.add_argument('-label_directory', default='D:\EmotionSegmentationData\CelebAMask-HQ\CelebAMask-HQ-mask-anno', type=str)
parser.add_argument('-image_directory', default='C:\home\laima\CelebAMask-HQ\CelebA-HQ-img', type=str)
parser.add_argument('-label_directory', default='C:\home\laima\CelebAMask-HQ\CelebAMask-HQ-mask-anno', type=str)
parser.add_argument('-data_len', default=30000, type=int)

args, other_args = parser.parse_known_args()

def rgb2gray(rgb):
    return np.dot(rgb[..., :3], [0.2989, 0.5870, 0.1140])

img_dir = args.image_directory
label_dir = args.label_directory

mmap_dir = args.datasource_directory

mmap_path = os.path.join(mmap_dir, 'data.mmap')
json_path = os.path.join(mmap_dir, 'data.json')
print(mmap_path)
if not os.path.exists(os.path.join(mmap_dir)):
    os.makedirs(os.path.join(mmap_dir))

data_len = args.data_len # 30000
# label_list = ['l_eye', 'r_eye']  # , 'l_eye', 'r_eye', 'l_brow', 'r_brow', 'mouth', 'u_lip', 'l_lip']
# label_list = ['eye', 'brow', 'mouth', 'lip']
label_list = ['eye', 'brow', 'mouth', 'lip', 'nose']

channel_list = ['photo_r', 'photo_g', 'photo_b', 'photo_grayscale'] + label_list
channels = len(channel_list)  # r, g, b, grayscale, label_list
width = 256
height = 256
data_shape = (data_len, channels, width, height)

# data description to json
data_desc = {'shape': data_shape,
             'channels': channel_list}

with open(json_path, "w+") as a_file:
    json.dump(data_desc, a_file)

fp = np.memmap(mmap_path, dtype='float16', mode='w+', shape=data_shape)
for i in tqdm(range(data_len)):
    # if i % 100 == 0:
    #     print(i)

    img_path = os.path.join(img_dir, str(i) + '.jpg')
    image = Image.open(img_path)
    image_transform = image.resize((width, height))
    np_image = np.asarray(image_transform)

    # RGB image
    channel_r = channel_list.index('photo_r')
    channel_b = channel_list.index('photo_b')
    np_image = np.swapaxes(np_image, 0, 2)
    # print(np_image.shape)
    fp[i, channel_r:channel_b + 1, :, :] = np_image

    # grayscale image
    image_grayscale = np.asarray(image_transform.convert("L"))  #
    # normalizing image
    mu = np.mean(image_grayscale)
    std = np.std(image_grayscale)
    image_grayscale = (image_grayscale - mu) / std
    channel_grayscale = channel_list.index('photo_grayscale')
    fp[i, channel_grayscale:channel_grayscale + 1, :, :] = image_grayscale[:]

    # masks
    folder_num = i // 2000
    for idx, label in enumerate(label_list):
        file_prefix = str(i).rjust(5, '0')
        #list of labels that will be united into one (e.g l_eye and r_eye)
        label_path_list = glob.glob(f'{label_dir}/{folder_num}/{file_prefix}_*{label}.png')
        mask = np.zeros((width, height))
        for label_path in label_path_list:
            if os.path.exists(label_path):
                # print(label, idx + 1)
                image_label = Image.open(label_path).convert('P')
                image_label_transform = image_label.resize((width, height))
                np_label_image = np.asarray(image_label_transform)
                mask[np_label_image == 225] = 1
        channel_label = channel_list.index(label)
        fp[i, channel_label:channel_label + 1, :, :] = mask[:]

    fp.flush()
