import csv
from itertools import islice

import matplotlib.pyplot as plt
from PIL import Image
import numpy as np
import os.path
import cv2
# from torchvision import transforms
import json
import argparse
import glob
from tqdm import tqdm


def str_to_image(image_blob):
    ''' Convert a string blob to an image object. '''
    image_string = image_blob.split(' ')
    image_data = np.asarray(image_string, dtype=np.uint8).reshape(48, 48)
    return image_data  # Image.fromarray(image_data)


parser = argparse.ArgumentParser(add_help=False)
print(parser)

parser.add_argument('-datasource_directory', default='D:\\EmotionSegmentationData\\facial_expression_dataset', type=str)
# parser.add_argument('-datasource_directory', default='facial-expression-dataset', type=str)
# parser.add_argument('-data_file', default='facial-expression-dataset/icml_face_data.csv', type=str)
parser.add_argument('-data_file', default='facial-expression-dataset/fer2013/fer2013.csv', type=str)
# parser.add_argument('-label_directory', default='C:\home\laima\CelebAMask-HQ\\CelebAMask-HQ-mask-anno', type=str)
# parser.add_argument('-data_len', default=100, type=int)  # 35885
parser.add_argument('-data_len', default=35885, type=int)  # 35885

args, other_args = parser.parse_known_args()

data_path = args.data_file

mmap_dir = args.datasource_directory
mmap_path = os.path.join(mmap_dir, 'data.mmap')
json_path = os.path.join(mmap_dir, 'data.json')
label_path = os.path.join(mmap_dir, 'fer_label.csv')
# label_path = os.path.join(mmap_dir, 'fer_label.mmap')
print(mmap_path)
if not os.path.exists(os.path.join(mmap_dir)):
    os.makedirs(os.path.join(mmap_dir))

data_len = args.data_len

channel_list = ['photo_grayscale']
channels = len(channel_list)  # r, g, b, grayscale, label_list
width = 48
height = 48
data_shape = (data_len, channels, width, height)
# data description to json
data_desc = {'shape': data_shape,
             'channels': channel_list}

with open(json_path, "w+") as a_file:
    json.dump(data_desc, a_file)

fer_emotion_dict = {0: 'Angry',
                    1: 'Disgust',
                    2: 'Fear',
                    3: 'Happy',
                    4: 'Sad',
                    5: 'Surprise',
                    6: 'Neutral'}  # (0=Angry, 1=Disgust, 2=Fear, 3=Happy, 4=Sad, 5=Surprise, 6=Neutral)
fer_dict_path = os.path.join(mmap_dir, 'fer_emotion_dict.json')
with open(fer_dict_path, "w+") as a_file:
    json.dump(fer_emotion_dict, a_file)


fp = np.memmap(mmap_path, dtype='float16', mode='w+', shape=data_shape)
labelfile = open(label_path, 'w', newline='')
writer = csv.writer(labelfile)
with open(data_path, 'r') as csvfile:
    fer_rows = csv.reader(csvfile, delimiter=',')  # emotion, pixels, usage
    for idx_row, row in tqdm(enumerate(islice(fer_rows, 1, data_len + 1))):
        writer.writerow([row[0], row[2]])
        print(idx_row, row[0], row[2])
        # if idx_row > 0:
        image_grayscale = str_to_image(row[1])
        mu = np.mean(image_grayscale)
        std = np.std(image_grayscale)
        image_grayscale = (image_grayscale - mu) / std
        fp[idx_row, :, :, :] = image_grayscale
        fp.flush()
labelfile.close()
# print(fp)

# with open(label_path,'r') as csvfile:
#     fer_label = csv.reader(csvfile)
#     for row in fer_label:
#         print(row)
