import glob
import json
import os.path
import shutil
from datetime import datetime
from time import time
import matplotlib
import pandas as pd
import torch
import torch.nn.functional as F
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.patches import Patch
from torch.nn.utils.rnn import PackedSequence
import argparse
import torch.utils.data
from tqdm import tqdm

from facial_expressions_utilities import filter_masks
from modules.file_utils import FileUtils
from modules.facial_expression_dataset import FacialExpressionDataset
from modules.celeb_dataset import CelebDataset

plt.rcParams["figure.figsize"] = (12, 9)
import cv2


class DictToObj:
    def __init__(self, **entries):
        self.__dict__.update(entries)


parser = argparse.ArgumentParser(add_help=False)

parser.add_argument('-datasource_directory', default='./ferplus_masks_dataset_withoutglasses', type=str)
# parser.add_argument('-datasource_directory', default='D:\\EmotionSegmentationData\\ferplus_masks_dataset', type=str)
# parser.add_argument('-data_preprocess_directory', default='D:\\EmotionSegmentationData\\facial_expression_dataset', type=str)
parser.add_argument('-data_preprocess_directory', default='./facial-expression-dataset', type=str)
parser.add_argument('-mmap_filename', default='data.dat', type=str)
parser.add_argument('-json_filename', default='data.json', type=str)
parser.add_argument('-fer_dict_filename', default='fer_emotion_dict.json', type=str)

parser.add_argument('-results_foldername', default='results', type=str)
parser.add_argument('-sequence_name', default='segmentation_classes', type=str)
# parser.add_argument('-run_name', default='test_model_augmentation_43-1-run-21-11-08--08-08-01-711214', type=str)
# parser.add_argument('-run_name', default='test_model_augmentation_fcn-2-run-21-11-08--08-15-24-255000', type=str)
# parser.add_argument('-run_name', default='test_model_types_deeplabv3-1-run-21-11-09-10-18-40-21-11-09--10-18-43-027433', type=str)
# parser.add_argument('-run_name', default='test_model_augmentation_44-21-11-11-13-43-50-1-run-21-11-11--13-43-53-410492',
# type=str)
parser.add_argument('-run_name', default='segmentation_classes-4-run-21-12-12-10-42-16', type=str)
# parser.add_argument('-run_name', default='segmentation-3-run-21-11-21-19-38-21', type=str)
# parser.add_argument('-run_name', default='segmentation-7-run-21-11-24-00-42-41', type=str)
# parser.add_argument('-run_name', default='segmentation-2-run-21-11-21-19-43-41', type=str)
parser.add_argument('-preprocess', default=True, type=lambda x: (str(x).lower() == 'true'))

args, other_args = parser.parse_known_args()

run_path = f'{args.results_foldername}/{args.sequence_name}/{args.run_name}'
# run_path = f'{args.results_foldername}/{args.run_name}'
print(run_path)

model_list = glob.glob(f'{run_path}/model/*.pt')
model_list.sort(key=os.path.getmtime)
print(model_list)
args.path_to_model = model_list[-1]
print(f'path_to_model: {args.path_to_model}')

path_to_args_json = f'{run_path}/args.json'
args.path_to_args_json = path_to_args_json

# figure_masks_path = f"{args.datasource_directory}/figures_masks_{datetime.utcnow().strftime(f'%y-%m-%d--%H-%M-%S')}"
# figure_masks_path = f"{args.datasource_directory}/{args.run_name}"
figure_masks_path = f"{args.datasource_directory}"
if not os.path.exists(figure_masks_path):
    os.makedirs(figure_masks_path)

# copy model args to figures folder
shutil.copy2(args.path_to_args_json, figure_masks_path)

with open(args.path_to_args_json, 'r') as fp:
    args_model = json.load(fp)
    args_model = DictToObj(**args_model)

# emotion_name_path = os.path.join(args.data_preprocess_directory, args.fer_dict_filename)
# with open(emotion_name_path, 'r') as fp:
#     emotion_names = json.load(fp)
# print(emotion_names)

print(args_model)

# path_to_data_json = f'{run_path}/data.json'
# with open(path_to_data_json, 'r') as fp:
#     data_params = json.load(fp)
mask_name_list = args_model.output_channels
classes = len(mask_name_list)
Model = getattr(__import__('models.' + args_model.model, fromlist=['Model']), 'Model')
model = Model(classes=classes, args=args_model)  # .to(args_model.device)

# original saved file with DataParallel
state_dict = torch.load(args.path_to_model)

# create new OrderedDict that does not contain module.
from collections import OrderedDict

# new_state_dict = OrderedDict()
# for k, v in state_dict.items():
#     name = k[7:]  # remove module.
#     new_state_dict[name] = v
#
# # load params
# model.load_state_dict(new_state_dict)

# model.load_state_dict(torch.load(args.path_to_model, map_location=DEVICE))

# # model = Model(args_model)
device = torch.device('cpu')
model.load_state_dict(torch.load(args.path_to_model, map_location=device))
# # model.to(args_model.device)


dataset_full = FacialExpressionDataset(args)  # CelebDataset(args) #

data_loader = torch.utils.data.DataLoader(
    dataset=dataset_full,
    batch_size=1,
    shuffle=False,
    drop_last=False
)

emotion_names = dataset_full.getEmotionDict()

# plt.figure()
# col_number = 6
# row_number = 1

mmap_dir = figure_masks_path  # args.datasource_directory
# if not os.path.exists(os.path.join(mmap_dir)):
#     os.makedirs(os.path.join(mmap_dir))

mmap_path = os.path.join(mmap_dir, 'data.mmap')
json_path = os.path.join(mmap_dir, 'data.json')
label_path = os.path.join(mmap_dir, 'fer2013new.csv')
mmap_not_included_path = os.path.join(mmap_dir, 'data_not_included.mmap')
label_not_included_path = os.path.join(mmap_dir, 'fer2013new_not_included.csv')

mask_name_list = args_model.output_channels
print(mask_name_list)
cmap_list = ['Blues', 'Greens', 'Purples', 'Reds', 'Oranges']

channel_list = args_model.input_channels + mask_name_list[:-1]
channels = len(channel_list)  # r, g, b, grayscale, label_list
width = 256
height = 256
data_shape = (len(dataset_full), channels, width, height)
print(data_shape)
# data description to json
data_desc = {'shape': data_shape,
             'channels': channel_list}

with open(json_path, "w+") as a_file:
    json.dump(data_desc, a_file)

fp = np.memmap(mmap_path, dtype='float16', mode='w+', shape=data_shape)

fp_not_included = np.memmap(mmap_not_included_path, dtype='float16', mode='w+', shape=data_shape)
#
# if label_path:
#     labels_df = pd.read_csv(label_path, usecols=emotion_names.values())
#     labels_df_not_included = pd.read_csv(label_not_included_path, usecols=emotion_names.values())
# else:
#     labels_df = pd.DataFrame(columns=emotion_names.values())
#     labels_df_not_included = pd.DataFrame(columns=emotion_names.values())
labels_df = pd.DataFrame(columns=emotion_names.values())
labels_df_not_included = pd.DataFrame(columns=emotion_names.values())
print(labels_df)
# fig_number = 100
idx_included = 0
idx_not_included = 0
for idx, (x, y, z) in tqdm(enumerate(data_loader)):

    # if idx < 29222:
    #     continue

    emotion_prob_str = ''
    z_item_max = -1
    z_idx_max = 0
    for z_idx, z_item in enumerate(z[0]):
        if z_item > 0:
            # print(z_item)
            # print(z_idx, emotion_names[int(z_idx)], z_item)
            emotion_prob_str += f'{emotion_names[int(z_idx)]}: {z_item} '
        if z_item > z_item_max:
            z_item_max = z_item
            z_idx_max = z_idx
            # print(f'max_idx: {z_idx_max}')
    emotion_max = emotion_names[int(z_idx_max)]

    # fp[idx, :1, :, :] = x[0, :, :, :]

    # x = x.to(args_model.device)
    model = model.eval()
    with torch.no_grad():
        y_prim = model.forward(x)
        # y_prim = y_prim.cpu()
        # x = x.cpu()
    masks = y_prim[0, :-1, :, :]

    # include_masks = include_image(masks, mask_name_list, emotion_max, min_pixels=10)
    filtered_masks, include_masks = filter_masks(masks, emotion_name=emotion_max, min_pixels=30)
    if include_masks:
        savefolder = 'included_data'
        fp[idx_included, :1, :, :] = x[0, :, :, :]
        fp[idx_included, 1:, :, :] = filtered_masks
        labels_df.loc[idx_included, :] = z.numpy()
        # print(labels_df)

        idx_included += 1
    else:
        savefolder = 'not_included_data'
        fp_not_included[idx_not_included, :1, :, :] = x[0, :, :, :]
        fp_not_included[idx_not_included, 1:, :, :] = filtered_masks
        labels_df_not_included.loc[idx_included, :] = z.numpy()
        idx_not_included += 1

    fp.flush()
    fp_not_included.flush()
    labels_df.to_csv(label_path)
    labels_df_not_included.to_csv(label_not_included_path)

    # plt.clf()
    # plt.figure(figsize=(12, 9))
    # # plt.suptitle(f'{y[0]} {emotion_names[str(y.numpy()[0])]}')
    #
    # plt.suptitle(f'{emotion_prob_str}, max: {emotion_max} ({z_item_max})')
    # # plt.suptitle(f'{y[0]} {emotion_names[int(y.numpy()[0])]}')
    # row_number = 3
    # col_number = len(channel_list) - 1
    # # print(row_number, col_number)
    # # plt.subplot(row_number, col_number, 1)
    # ax1 = plt.subplot(row_number, col_number, 1)
    # ax1.imshow(x[0, 0], cmap=plt.get_cmap('gray'))
    # ax2 = plt.subplot(row_number, col_number, 2)
    # ax2.imshow(x[0, 0], cmap=plt.get_cmap('gray'))
    #
    # ax3 = plt.subplot(row_number, col_number, 3)
    # ax4 = plt.subplot(row_number, col_number, 4)
    # legend_elements = []
    # # masked_list = []
    # for idx_mask in range(len(mask_name_list) - 1):
    #     cmap = matplotlib.cm.get_cmap(cmap_list[idx_mask])
    #     rgba = cmap(1.)
    #     legend_elements.append(Patch(facecolor=rgba))
    #     masked = np.ma.masked_where(filtered_masks[idx_mask, :, :] == 0, filtered_masks[idx_mask, :, :])
    #     masked = np.expand_dims(masked, axis=2)
    #     # [masked * color for color in rgba]
    #     masked_rgb = np.concatenate([masked * color for color in rgba], axis=2)
    #
    #     ax2.imshow(masked_rgb, alpha=0.4)
    #     ax3.imshow(masked_rgb)
    #     ax = plt.subplot(row_number, col_number, col_number + 1 + idx_mask)
    #     ax.imshow(y_prim[0, idx_mask, :, :], cmap=cmap)
    #
    #     # item_mask = y_prim[0, idx_mask, :, :]
    #     # item_mask_np = np.array((1 - item_mask) * 255, dtype='uint8')
    #     # print(item_mask_np.min(), item_mask_np.max(), item_mask_np.shape)
    #     # print(type(item_mask), type(item_mask_np), type(item_mask_np[0, 0]))
    #     # orb = cv2.ORB_create()
    #     # keypoints = detector.detect(item_mask_np)
    #     # # print(len(keypoints), keypoints)
    #     # im_with_keypoints = cv2.drawKeypoints(item_mask_np, keypoints, np.array([]), (0, 0, 255),
    #     #                                       cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
    #     # ax.imshow(im_with_keypoints)
    #
    #     ax0 = plt.subplot(row_number, col_number, col_number * 2 + 1 + idx_mask)
    #     ax0.imshow(masked_rgb, cmap=cmap)
    #     # ax0.imshow(filtered_masks[idx_mask, :, :], cmap=cmap)
    #     ax0.set_xlabel(mask_name_list[idx_mask])
    #     if idx_mask == 0:
    #         ax0.set_ylabel('filterd masks')
    #         ax.set_ylabel('original masks')
    #
    # ax4.legend(legend_elements, mask_name_list)  # , ncol=2)
    #
    # plt.tight_layout()
    #
    # filename = f'{idx:05d}.png'
    # savepath = f"{figure_masks_path}/{savefolder}"
    # FileUtils.createDir(savepath)
    # plt.savefig(f'{savepath}/{filename}', bbox_inches='tight')
    # # print(filename)
    # plt.close()

# new shape based on included data
data_shape = (idx_included, channels, width, height)

# data description to json
data_desc = {'shape': data_shape,
             'channels': channel_list}

with open(json_path, "w+") as a_file:
    json.dump(data_desc, a_file)

fp_reshape = np.memmap(mmap_path, dtype='float16', mode='r', shape=data_shape)
fp_reshape.flush()

fp_not_included_reshape = np.memmap(mmap_not_included_path, dtype='float16', mode='r',
                                    shape=(idx_not_included, channels, width, height))
fp_not_included_reshape.flush()

# plt.show()
