import json
import os

import cv2
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch


def filter_masks(masks_init, emotion_name=None, min_pixels=1):
    include_masks = True

    if emotion_name == 'NF':
        include_masks = False
        return masks_init, include_masks

    masks = np.zeros_like(masks_init)
    masks[masks_init < 0.5] = 0
    masks[masks_init >= 0.5] = 1

    shape = masks_init.shape
    lip_mouth = np.zeros([shape[-1], shape[-1]])
    lip_mouth[masks_init[2] > 0.5] = 1
    lip_mouth[masks_init[3] > 0.5] = 1
    lip_mouth = np.expand_dims(lip_mouth, axis=0)
    masks = np.concatenate([lip_mouth, masks], axis=0)

    filtered_masks = masks.copy()
    two_eyes = True
    two_brows = True
    for idx_mask, item_mask in enumerate(masks):
        # item_mask_np = np.array(item_mask, dtype='uint8')
        # item_mask_np[0, :] = 0  # robeža, lai var atrast pareizas kontūras
        # item_mask_np[255, :] = 0
        # item_mask_np[:, 0] = 0
        # item_mask_np[:, 255] = 0

        item_mask_np = np.array((1 - item_mask) * 255, dtype='uint8')
        item_mask_np[0, :] = 255
        item_mask_np[255, :] = 255
        item_mask_np[:, 0] = 255
        item_mask_np[:, 255] = 255

        ret, thresh = cv2.threshold(item_mask_np, 187, 255, 0)

        contours_all, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        # print(contours_all)
        contours = contours_all[1:]  # bez ārējās kontūras
        all_areas = [cv2.contourArea(c) for c in contours]
        # print(idx_mask, all_areas)

        if idx_mask == 0 and all_areas:
            pixel_limit = max(all_areas)  # lūpas un muti skatās kopā, un tiek vaļā no visa pārējā
        elif (idx_mask in [1,2,4]) and not all_areas:
            # filtered_masks = masks
            include_masks = False
            # print(idx_mask, all_areas)
            break
        else:
            pixel_limit = min_pixels

        area_count = 0
        for idx_con, item_con in enumerate(contours):
            area = all_areas[idx_con]
            if area > 0:
                x = np.arange(shape[-1])
                y = np.arange(shape[-1])
                xv, yv = np.meshgrid(x, y, indexing='xy')
                points = np.hstack((xv.reshape((-1, 1)), yv.reshape((-1, 1))))
                path = matplotlib.path.Path(item_con.reshape(-1, 2))
                filled_con = path.contains_points(points)
                filled_con.shape = xv.shape

                if area < pixel_limit:
                    filtered_masks[idx_mask, filled_con == 1] = 0
                    if idx_mask == 0:  # ja ir kopēja maska lūpas + mute, tad lieko atņemt arī atsevišķajās maskās
                        filtered_masks[3, filled_con == 1] = 0
                        filtered_masks[4, filled_con == 1] = 0
                else:
                    area_count += 1

        if (idx_mask in [1, 2, 4]) and (area_count == 0):
            include_masks = False
            # print(idx_mask, area_count)
            break
        elif (idx_mask == 1) and (area_count < 2):
            two_eyes = False
            if not two_brows:
                include_masks = False
                break
        elif (idx_mask == 2) and (area_count < 2):
            two_brows = False
            if not two_eyes:
                include_masks = False
                break



    filtered_masks = filtered_masks[1:, :, :]
    return filtered_masks, include_masks


def include_image(masks, mask_name_list, emotion_name, min_pixels=1):
    include_masks = True
    sum_mask = torch.sum(masks, dim=(1, 2))

    if emotion_name == 'NF':
        include_masks = False
        return include_masks

    for idx_mask, mask_name in enumerate(mask_name_list):
        if mask_name in ['mouth', 'other']:
            continue
        if sum_mask[idx_mask] <= min_pixels:
            include_masks = False
            break

    return include_masks

if __name__ == '__main__':
    mmap_dir = 'ferplus_masks_dataset_withoutglasses/segmentation-3-run-21-11-21-19-38-21'
    json_path = os.path.join(mmap_dir, 'data.json')
    with open(json_path) as fp:
        parameters = json.load(fp)

    shape = tuple(parameters['shape'])
    channels = parameters['channels']

    mmap_path = os.path.join(mmap_dir, 'data.mmap')
    mmap = np.memmap(mmap_path, dtype='float16', mode='r', shape=shape)

    # idx = 0
    for idx in range(10):
        img = torch.FloatTensor(np.array(mmap[idx, :, :, :]))
        masks = img[1:, :, :]

        filtered_masks, include_masks = filter_masks(masks, min_pixels=50)

        print(idx, include_masks)
        row_count = 2
        col_count = 4
        plt.figure(figsize=(8, 3))
        for idx_mask, item_mask in enumerate(filtered_masks):
            plt.subplot(row_count, col_count, 1 + idx_mask)
            plt.imshow(masks[idx_mask,:,:], alpha=1)

            plt.subplot(row_count, col_count, 5 + idx_mask)
            plt.imshow(item_mask, alpha=1)

    plt.show()
    # print(filtered_masks, include_masks)
