import hashlib
import math
import pickle
import shutil
from datetime import datetime
import random
from time import time
# from pytorch_model_summary import summary
import matplotlib
import matplotlib.pyplot as plt
# from matplotlib.ticker import FormatStrFormatter
import numpy as np
import pandas as pd
import sklearn.metrics
# from sklearn.metrics import f1_score
import torch.utils.data
import argparse
import os
import json
import torch.nn.functional as F
import torchvision.transforms as T
from matplotlib.patches import Patch

from tqdm import tqdm
from matplotlib.ticker import FormatStrFormatter

plt.rcParams["figure.figsize"] = (18, 5)
# plt.rcParams["figure.figsize"] = (24,10)

from modules.file_utils import FileUtils
from modules.csv_utils_2 import CsvUtils2
from modules.celeb_dataset import CelebDataset
from modules.metric_functions import IoU, dice_coeficient, make_figures
from modules.facial_expression_dataset import FacialExpressionDataset, FacialExpressionMasksDataset
import logging

parser = argparse.ArgumentParser(add_help=False)
print(parser)

# parser.add_argument('-datasource_directory', default='./celeb_hq_dataset', type=str)
parser.add_argument('-datasource_directory', default='./ferplus_masks_dataset_withoutglasses', type=str)
# parser.add_argument('-datasource_directory', default='./ferplus_masks_dataset', type=str)
parser.add_argument('-mmap_filename', default='data.mmap', type=str)
parser.add_argument('-json_filename', default='data.json', type=str)
parser.add_argument('-fer_dict_filename', default='fer_emotion_dict.json', type=str)
parser.add_argument('-learning_rate', default=1e-3, type=float)
parser.add_argument('-batch_size', default=8, type=int)
parser.add_argument('-epochs', default=100, type=int)
parser.add_argument('-sequence_name', default='sequence1', type=str)
# parser.add_argument('-run_name', default='segmentation-3-run-21-11-21-19-38-21', type=str)
parser.add_argument('-run_name', default='run', type=str)
# parser.add_argument('-is_cuda', default=True, type=lambda x: (str(x).lower() == 'true'))
parser.add_argument('-is_cuda', default=False, type=lambda x: (str(x).lower() == 'true'))
# parser.add_argument('-model', default='2_model_pre_trained_fcn', type=str)
parser.add_argument('-model', default='7_model_pre_trained_densenet', type=str)
parser.add_argument('-preprocess', default=False, type=lambda x: (str(x).lower() == 'true'))
parser.add_argument('-class_num', default=10, type=int)
parser.add_argument('-in_image_channels', default=4, type=int)
parser.add_argument('-early_stopping_patience', default=10, type=int)

args, other_args = parser.parse_known_args()
print(args)

# MAX_LEN = 90


if not torch.cuda.is_available() or not args.is_cuda:
    args.device = 'cpu'
    args.is_cuda = False
    print('cuda not available')
else:
    args.device = 'cuda'
    print(f'cuda devices: {torch.cuda.device_count()}')

# Begin code
path_sequence = f'./results/{args.sequence_name}'
path_run = f'{path_sequence}/{args.run_name}'
path_run_fig = f'{path_run}/figures'
path_modelsave = f'{path_run}/model'
dir_cache = './cache'

# FileUtils.createDir(path_run)
FileUtils.createDir(path_run_fig)
FileUtils.createDir(path_modelsave)
FileUtils.createDir(dir_cache)

CsvUtils2.create_global(path_sequence)
CsvUtils2.create_local(path_sequence, args.run_name)

# copy data.json to results folder
shutil.copy2(f'{args.datasource_directory}/{args.json_filename}', path_run)
print('csv utils setup done')

rootLogger = logging.getLogger()
logFormatter = logging.Formatter("%(asctime)s [%(process)d] [%(thread)d] [%(levelname)s]  %(message)s")
rootLogger.level = logging.DEBUG  # level

fileHandler = logging.FileHandler(f'{path_run}/log_main.txt')
fileHandler.setFormatter(logFormatter)
rootLogger.addHandler(fileHandler)

consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)
rootLogger.addHandler(consoleHandler)

dataset_full = FacialExpressionMasksDataset(args)
len_dataset = len(dataset_full)
emotion_names_full = dataset_full.getEmotionDict()
# args.class_num = 4
print(emotion_names_full)
# if args.class_num == 4:
#     class_idx_list = [0, 1, 3, 4]
# else:
#     class_idx_list = range(args.class_num)
class_idx_list = range(args.class_num)
emotion_names = {k: emotion_names_full[k] for k in class_idx_list if k in emotion_names_full}
print(emotion_names)
# input()

# print(len_dataset)
logging.info(f'Dataset length: {len_dataset}')
labels_df = dataset_full.fer_labels_new_df.copy()
labels_df['max_emotion'] = labels_df.idxmax(axis=1)
logging.info(f'Labels: {labels_df}')

emo_sample_count = np.zeros(args.class_num)
for idx_count, idx_col in enumerate(class_idx_list):
    count = labels_df[labels_df['max_emotion'] == labels_df.columns[idx_col]]['max_emotion'].count()
    emo_sample_count[idx_count] = count
    logging.info(f'{labels_df.columns[idx_col]} count: {count}')
totalcount = np.sum(emo_sample_count[np.nonzero(emo_sample_count)])
test_count0 = int(totalcount * 0.2 / args.class_num)
logging.info(f'totalcount: {totalcount}')
logging.info(f'test_count0: {test_count0}')
mincount = np.min(emo_sample_count[np.nonzero(emo_sample_count)])
logging.info(f'mincount: {mincount}')
test_count = int(mincount / 2)
if test_count > test_count0:
    test_count = test_count0
logging.info(f'test_count: {test_count}')
# input()
args.max_len_per_class = test_count

test_idx_list = []
train_idx_list = []
train_samples_in_class = []
str_args = [str(it) for it in [args.max_len_per_class, args.class_num]]
hash_args = hashlib.md5((''.join(str_args)).encode()).hexdigest()
path_cache = f'{dir_cache}/{hash_args}.pkl'
if os.path.exists(path_cache):
    print('loading from cache')
    with open(path_cache, 'rb') as fp:
        train_idx_list, test_idx_list, train_samples_in_class = pickle.load(fp)
else:
    print('generating new indexes')
    emo_sample_count = []
    for idx_col in class_idx_list:
        emo_idx_list = labels_df[labels_df['max_emotion'] == labels_df.columns[idx_col]].index
        if list(emo_idx_list):
            test_idx_list += list(emo_idx_list[:test_count])
            train_idx_list += list(emo_idx_list[test_count:])
            train_samples_in_class.append(len(emo_idx_list[test_count:]))
            emo_sample_count.append(len(emo_idx_list))
        else:
            train_samples_in_class.append(0)
            emo_sample_count.append(0)
    with open(path_cache, 'wb') as fp:
        pickle.dump((train_idx_list, test_idx_list, train_samples_in_class), fp)

logging.info(f'test_idx_list: {test_idx_list}')
logging.info(f'train_idx_list: {train_idx_list}')
logging.info(f'train_samples_in_class: {train_samples_in_class}')
logging.info(f'emo_sample_count: {emo_sample_count}')
logging.info(f'channels: {dataset_full.channels}')

for idx_emo, emo_name in emotion_names.items():
    print(idx_emo, emo_name)
# input()


random.seed(0)
random.shuffle(train_idx_list)
random.shuffle(test_idx_list)
random.seed()

dataset_train = torch.utils.data.Subset(dataset_full, train_idx_list)
dataset_test = torch.utils.data.Subset(dataset_full, test_idx_list)

# data loaders
data_loader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=args.batch_size,
    shuffle=True,
    drop_last=(len(dataset_train) % args.batch_size == 1)
)
data_loader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=args.batch_size,
    shuffle=False,
    drop_last=(len(dataset_test) % args.batch_size == 1)
)

Model = getattr(__import__('models.' + args.model, fromlist=['Model']), 'Model')
model = Model(CLASS_NUM=args.class_num, in_image_channels=args.in_image_channels).to(args.device)
weights = torch.FloatTensor(1 - (train_samples_in_class / np.sum(train_samples_in_class)))
weights = weights.to(args.device)
logging.info(f'loss weights: {weights}')
loss_function = torch.nn.CrossEntropyLoss(weight=weights)
loss_function_test = torch.nn.CrossEntropyLoss()

is_data_parallel = False
if torch.cuda.device_count() > 1:
    is_data_parallel = True
    model = torch.nn.DataParallel(model, dim=0)

optimizer = torch.optim.Adam(model.parameters(), lr=args.learning_rate)

FileUtils.writeJSON(f'{path_run}/args.json', args.__dict__)
metrics = {}
metrics_mean_dict = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        'acc',
        'f1_score'
    ]:
        metrics[f'{stage}_{metric}'] = []
        metrics_mean_dict[f'{stage}_{metric}'] = 1e20

for idx_emo, emo_name in emotion_names.items():
    metrics[f'test_f1_score_{emo_name}'] = []
    metrics_mean_dict[f'test_f1_score_{emo_name}'] = 1e20

best_test_loss = 1e20
best_test_f1_score = 0
best_test_acc = 0
best_test_counter = 0
cmap_list = ['Blues', 'Greens', 'Purples', 'Reds', 'Oranges']

for epoch in range(1, args.epochs):
    for data_loader in [data_loader_train, data_loader_test]:
        metrics_epoch = {key: [] for key in metrics.keys()}
        stage = 'train'
        if data_loader == data_loader_test:
            stage = 'test'

        for input_masks, label_max, label_votes, grayscale in data_loader:
            # i1 = 0
            # for x1, y1, z1, img1 in data_loader:
            #
            #     if i1 > 0:
            #         break
            #     i1 += 1
            #     x, y, z, img = x1, y1, z1, img1

            y = label_max
            z = label_votes
            if args.in_image_channels == 1:
                x = grayscale
            else:
                x = input_masks

            x = x.to(args.device)
            # y = torch.argmax(z, dim=1)

            y = y.to(device=args.device, dtype=torch.long)
            # print(z)
            if stage == 'test':
                model = model.eval()
                with torch.no_grad():
                    y_prim = model.forward(x)

                    loss = loss_function_test(y_prim, y)
            else:
                model = model.train()
                y_prim = model.forward(x)
                # print(y, y_prim)
                loss = loss_function(y_prim, y)
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

            loss = loss.cpu()
            y_prim = y_prim.cpu()
            x = x.cpu()
            y = y.cpu()

            np_x = x.data.numpy()
            np_y_prim = y_prim.data.numpy()
            np_y = y.data.numpy()
            np_z = z.data.numpy()

            idx_y = np.argmax(np_z, axis=1)
            idx_y_prim = np.argmax(np_y_prim, axis=1)

            acc = np.average((idx_y == idx_y_prim) * 1.0)
            f1_score = sklearn.metrics.f1_score(idx_y, idx_y_prim, average='weighted')  # , sample_weight=weights)
            metrics_epoch[f'{stage}_loss'].append(loss.item())
            metrics_epoch[f'{stage}_acc'].append(acc)
            metrics_epoch[f'{stage}_f1_score'].append(f1_score)

            # slice idx_y for each idx_class value and calculate f1_score for each class
            for idx_emo, emo_name in emotion_names.items():
                i_list = (idx_y == idx_emo).nonzero()[0]
                # print(i_list, idx_y[i_list], idx_y_prim[i_list])
                if len(i_list) > 0:
                    idx_y_emo = idx_y[i_list]
                    idx_y_prim_emo = idx_y_prim[i_list]
                    f1_score_emo = sklearn.metrics.f1_score(idx_y_emo, idx_y_prim_emo, average='weighted')
                    # print(i_list, idx_y_emo, idx_y_prim_emo, f1_score_emo)
                    metrics_epoch[f'test_f1_score_{emo_name}'].append(f1_score_emo)
            # if stage == 'test':
            #     metrics[f'test_f1_score_{emotion_names[idx_y]}_{idx_y}'].append(f1_score)

        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')
                metrics_mean_dict[key] = value

    best_test_counter += 1
    if metrics_mean_dict['test_loss'] < best_test_loss:
        best_test_counter = 0
        best_test_loss = metrics_mean_dict['test_loss']
        # best_test_acc = metrics_mean_dict['test_acc']
        # best_test_f1_score = metrics_mean_dict['test_f1_score']
        metrics_mean_dict['best_test_loss'] = best_test_loss
        metrics_mean_dict['best_test_acc'] = metrics_mean_dict['test_acc']
        metrics_mean_dict['best_test_f1_score'] = metrics_mean_dict['test_f1_score']
        metrics_strs.append(f'best_test_loss: {round(best_test_loss, 2)}')
        metrics_strs.append(f'best_test_acc: {round(metrics_mean_dict["test_acc"], 2)}')
        metrics_strs.append(f'best_test_f1_score: {round(metrics_mean_dict["test_f1_score"], 2)}')
        for idx_emo, emo_name in emotion_names.items():
            metrics_mean_dict[f'best_test_f1_score_{emo_name}'] = metrics_mean_dict[f'test_f1_score_{emo_name}']
            metrics_strs.append(
                f'best_test_f1_score_{emo_name}: {round(metrics_mean_dict[f"test_f1_score_{emo_name}"], 2)}')

        #  save model parameters
        model_module = model
        if is_data_parallel:
            model_module = model.module
        torch.save(model_module.state_dict(), f'{path_modelsave}/model-best.pt')

    print(f'epoch: {epoch} {" ".join(metrics_strs)}, counter: {best_test_counter}')
    # logging.info(f'epoch: {epoch} {" ".join(metrics_strs)}, counter: {best_test_counter}')

    CsvUtils2.add_hparams(
        path_sequence=path_sequence,
        run_name=args.run_name,
        args_dict=args.__dict__,
        metrics_dict=metrics_mean_dict,  # contains train + test metrics
        global_step=epoch
    )

    plt.clf()
    ax = plt.subplot(131)  # row col idx
    plts = []
    c = 0
    # plt.figure('Metrics')
    for key, value in metrics.items():
        plts += plt.plot(value, f'C{c}', label=key)
        # plt.xlim(0, args.epochs)
        ax.yaxis.set_major_formatter(FormatStrFormatter('%.3f'))
        ax = plt.twinx()

        c += 1
    plt.tick_params(
        axis='y',
        which='both',
        right=False,
        labelright=False
    )
    plt.legend(plts, [it.get_label() for it in plts])

    num_figs = 9 if grayscale.shape[0] >= 9 else grayscale.shape[0]

    for idx_fig, fig_nr in enumerate([4, 5, 6, 13, 14, 15, 22, 23, 24]):
        if idx_fig >= num_figs:
            break
        ax1 = plt.subplot(3, 9, fig_nr)
        color = 'green' if idx_y[idx_fig] == idx_y_prim[idx_fig] else 'red'
        ax1.set_title(
            f"y={idx_y[idx_fig]}, y'={idx_y_prim[idx_fig]}, {idx_y[idx_fig] == idx_y_prim[idx_fig]}\n{emotion_names[idx_y[idx_fig]]}\n{emotion_names[idx_y_prim[idx_fig]]}",
            c=color)
        ax1.imshow(grayscale[idx_fig, 0], cmap=plt.get_cmap('gray'))

        ax2 = plt.subplot(3, 9, fig_nr + 3)
        color = 'green' if idx_y[idx_fig] == idx_y_prim[idx_fig] else 'red'
        ax2.set_title(
            f"y={idx_y[idx_fig]}, y'={idx_y_prim[idx_fig]}, {idx_y[idx_fig] == idx_y_prim[idx_fig]}\n{emotion_names[idx_y[idx_fig]]}\n{emotion_names[idx_y_prim[idx_fig]]}",
            c=color)
        legend_elements = []
        for idx_mask in range(4):
            cmap = matplotlib.cm.get_cmap(cmap_list[idx_mask])
            rgba = cmap(1.)
            legend_elements.append(Patch(facecolor=rgba))
            masked = np.ma.masked_where(input_masks[idx_fig, idx_mask, :, :] == 0, input_masks[idx_fig, idx_mask, :, :])
            masked = np.expand_dims(masked, axis=2)
            masked_rgb = np.concatenate([masked * color for color in rgba], axis=2)
            ax2.imshow(masked_rgb)

    plt.tight_layout(pad=0.01)
    plt.savefig(f'{path_run_fig}/epoch{epoch}.png')
    plt.close()

    if best_test_counter > args.early_stopping_patience:
        print(best_test_counter, best_test_loss, metrics['test_loss'][-args.early_stopping_patience:])
        break
