import torch.nn.functional as F
import torch.utils.data
import math
import torchvision
from torchvision.models.segmentation.fcn import FCNHead


class Model(torch.nn.Module):
    def __init__(self, classes=2, args = None):
        super().__init__()
        # TODO make also model for fcn_resnet101 or enable by args
        self.whole_model = torchvision.models.segmentation.deeplabv3_resnet50(
            pretrained=True
        )
        self.backbone = self.whole_model.backbone
        in_channels = 2048
        inter_channels = in_channels // 4
        self.classifier = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels, inter_channels, kernel_size=3, padding=1, bias=False),
            torch.nn.BatchNorm2d(inter_channels),
            torch.nn.ReLU(),
            torch.nn.Dropout(0.1),
            torch.nn.Conv2d(inter_channels, classes, kernel_size=1)
        )

    def forward(self, x):
        x_3channels = x.expand(x.size(0), 3, x.size(2), x.size(3))
        z = self.backbone(x_3channels)

        z_prim = self.classifier.forward(z['out'])
        z_prim = F.interpolate(z_prim, size=x.shape[-2:], mode='bilinear', align_corners=False)

        y_prim = torch.softmax(z_prim, dim=1)
        return y_prim


# if __name__ == '__main__':
#     from torchinfo import summary
#
#     model = Model()
#     summary(model, input_size=(16,1,256,256))

