# import csv
# from itertools import islice
import pandas as pd

import torch.utils.data
import os
import json
import numpy as np
import torchvision.transforms as T


class FacialExpressionDataset(torch.utils.data.Dataset):
    def __init__(self, args):
        super(FacialExpressionDataset, self).__init__()
        # load mmap pointer
        if args.preprocess:
            mmap_dir = args.data_preprocess_directory
            self.idx_channels = 1
        else:
            mmap_dir = args.datasource_directory
            self.idx_channels = 1

        json_path = os.path.join(mmap_dir, 'data.json')
        with open(json_path) as fp:
            parameters = json.load(fp)

        self.shape = tuple(parameters['shape'])
        # length, chan_num, width, height = tuple(parameters['shape'])
        # self.shape = (length, chan_num - self.idx_channels, width, height)
        # print(self.shape)
        # self.shape[1] = self.shape[1] - self.idx_channels
        self.channels = parameters['channels'][self.idx_channels:]
        # print(self.channels)

        # label_path = os.path.join(mmap_dir, 'fer_label.csv')
        # # labelfile = open(label_path, 'r')
        # # reader = csv.reader(labelfile)
        # # self.fer_labels = islice(reader, 0, None)
        # self.fer_labels_df = pd.read_csv(label_path, header=None)
        # # print(self.fer_labels_df)

        label_path_new = os.path.join(mmap_dir, 'fer2013new.csv')
        self.labels = ['neutral', 'happiness', 'surprise', 'sadness', 'anger', 'disgust', 'fear', 'contempt', 'unknown',
                       'NF']
        self.fer_labels_new_df = pd.read_csv(label_path_new, header=0, skiprows=0, delimiter=',',
                                             usecols=self.labels)
        # print(self.fer_labels_new_df.columns)
        # print(self.fer_labels_new_df.head())
        self.emotion_dict = {}
        self.inverse_emotion_dict = {}
        for idx_col, col in enumerate(self.fer_labels_new_df.columns):
            self.emotion_dict[idx_col] = col
            self.inverse_emotion_dict[col] = idx_col
        # print(self.emotion_dict)
        # print(self.inverse_emotion_dict)

        mmap_path = os.path.join(mmap_dir, 'data.mmap')
        self.mmap = np.memmap(mmap_path, dtype='float16', mode='r', shape=self.shape)

    def __len__(self):
        return self.shape[0]  # get from json shape

    def __getitem__(self, idx):

        x = torch.FloatTensor(np.array(self.mmap[idx, :, :, :]))
        x = T.Resize(size=256)(x)

        # x = x_full[self.idx_channels:, :, :]

        # img = x_full[:self.idx_channels, :, :]

        z_series = self.fer_labels_new_df.iloc[idx, :]

        max_emotion = z_series[z_series == z_series.max()].index[0]
        # print(max_emotion)
        # print(self.inverse_emotion_dict[max_emotion])

        # print(z_series, type(z_series), z_series.max())
        #
        # print(z_series[z_series==z_series.max()].index[0])
        #
        z = np.array(z_series.values, dtype=int)
        #
        # maxvote = z.argmax()
        # print(f'maxvote: {maxvote}')

        emotion_idx = int(self.inverse_emotion_dict[max_emotion])
        # print(emotion_idx)

        # y = self.fer_labels_df.loc[idx, 0]

        # y = torch.FloatTensor([emotion_idx])
        # y = torch.FloatTensor([emotion_idx])
        z = torch.FloatTensor(z)
        y = torch.argmax(z)

        # print(y, z)

        # y = self.fer_labels_df.loc[idx, 0]
        return x, y, z

    def getEmotionDict(self):
        return self.emotion_dict


class FacialExpressionMasksDataset(torch.utils.data.Dataset):
    def __init__(self, args):
        super(FacialExpressionMasksDataset, self).__init__()
        # load mmap pointer
        if args.preprocess:
            mmap_dir = args.data_preprocess_directory
            self.idx_channels = 1
        else:
            mmap_dir = args.datasource_directory
            self.idx_channels = 1

        json_path = os.path.join(mmap_dir, 'data.json')
        with open(json_path) as fp:
            parameters = json.load(fp)

        self.shape = tuple(parameters['shape'])
        # length, chan_num, width, height = tuple(parameters['shape'])
        # self.shape = (length, chan_num - self.idx_channels, width, height)
        # print(self.shape)
        # self.shape[1] = self.shape[1] - self.idx_channels
        self.channels = parameters['channels'][self.idx_channels:]
        # print(self.channels)

        # label_path = os.path.join(mmap_dir, 'fer_label.csv')
        # # labelfile = open(label_path, 'r')
        # # reader = csv.reader(labelfile)
        # # self.fer_labels = islice(reader, 0, None)
        # self.fer_labels_df = pd.read_csv(label_path, header=None)
        # # print(self.fer_labels_df)

        label_path_new = os.path.join(mmap_dir, 'fer2013new.csv')
        self.fer_labels_new_df = pd.read_csv(label_path_new, header=0, skiprows=0, delimiter=',',
                                             usecols=['neutral', 'happiness', 'surprise', 'sadness', 'anger',
                                                      'disgust', 'fear', 'contempt', 'unknown', 'NF'])
        self.fer_labels_new_df = self.fer_labels_new_df[['neutral', 'happiness', 'sadness', 'anger', 'surprise',
                                                         'disgust', 'fear', 'contempt', 'unknown', 'NF']]
        # print(self.fer_labels_new_df.columns)
        # print(self.fer_labels_new_df.head())
        self.emotion_dict = {}
        self.inverse_emotion_dict = {}
        for idx_col, col in enumerate(self.fer_labels_new_df.columns):
            self.emotion_dict[idx_col] = col
            self.inverse_emotion_dict[col] = idx_col
        # print(self.emotion_dict)
        # print(self.inverse_emotion_dict)

        mmap_path = os.path.join(mmap_dir, 'data.mmap')
        self.mmap = np.memmap(mmap_path, dtype='float16', mode='r', shape=self.shape)

    def __len__(self):
        return self.shape[0]  # get from json shape

    def __getitem__(self, idx):

        x_full = torch.FloatTensor(np.array(self.mmap[idx, :, :, :]))
        x_full = T.Resize(size=256)(x_full)

        x = x_full[self.idx_channels:, :, :]

        img = x_full[:self.idx_channels, :, :]

        z_series = self.fer_labels_new_df.iloc[idx, :]

        max_emotion = z_series[z_series == z_series.max()].index[0]
        # print(max_emotion)
        # print(self.inverse_emotion_dict[max_emotion])

        # print(z_series, type(z_series), z_series.max())
        #
        # print(z_series[z_series==z_series.max()].index[0])
        #
        z = np.array(z_series.values, dtype=int)
        #
        # maxvote = z.argmax()
        # print(f'maxvote: {maxvote}')

        emotion_idx = int(self.inverse_emotion_dict[max_emotion])
        # print(emotion_idx)

        # y = self.fer_labels_df.loc[idx, 0]

        # y = torch.FloatTensor([emotion_idx])
        # y = torch.FloatTensor([emotion_idx])
        z = torch.FloatTensor(z)
        y = torch.argmax(z)

        # print(y, z)

        # y = self.fer_labels_df.loc[idx, 0]
        return x, y, z, img

    def getEmotionDict(self):
        return self.emotion_dict


if __name__ == '__main__':
    # import torch.utils.data
    import argparse

    parser = argparse.ArgumentParser(add_help=False)
    print(parser)

    parser.add_argument('-datasource_directory', default='../ferplus_masks_dataset_withoutglasses', type=str)
    parser.add_argument('-data_preprocess_directory', default='../facial-expression-dataset', type=str)
    parser.add_argument('-mmap_filename', default='data.dat', type=str)
    parser.add_argument('-json_filename', default='data.json', type=str)
    parser.add_argument('-fer_dict_filename', default='fer_emotion_dict.json', type=str)

    parser.add_argument('-results_foldername', default='.', type=str)
    parser.add_argument('-sequence_name', default='seq_results', type=str)
    # parser.add_argument('-run_name', default='test_model_augmentation_43-1-run-21-11-08--08-08-01-711214', type=str)
    # parser.add_argument('-run_name', default='test_model_augmentation_fcn-2-run-21-11-08--08-15-24-255000', type=str)
    # parser.add_argument('-run_name', default='test_model_types_deeplabv3-1-run-21-11-09-10-18-40-21-11-09--10-18-43-027433', type=str)
    parser.add_argument('-run_name',
                        default='test_model_augmentation_44-21-11-11-13-43-50-1-run-21-11-11--13-43-53-410492',
                        type=str)

    args, other_args = parser.parse_known_args()
    print(args)

    run_path = f'{args.results_foldername}/{args.sequence_name}/{args.run_name}'
    # run_path = f'{args.results_foldername}/{args.run_name}'
    print(run_path)

    path_to_args_json = f'{run_path}/args.json'
    args.path_to_args_json = path_to_args_json

    dataset_full = FacialExpressionDataset(args)
    print(dataset_full.shape)
