import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

sequence_name = 'segmentation_classes'
path = f'results/segmentation_classes/sequence-{sequence_name}.csv'

seq_df = pd.read_csv(path)

namelist = ['', '_eye', '_brow', '_mouth', '_lip', '_other']

cols = [f'best_test_iou{m}' for m in namelist]

print(seq_df[cols])

seq_df_sort = seq_df[seq_df['step']>35].sort_values('best_test_iou', ascending=False)
seq_df_sort.reset_index(inplace=True)
# Number of rows to drop
n = 4
# Dropping last n rows using drop
# seq_df_sort.drop(seq_df_sort.tail(n).index, inplace=True)
# seq_df_sort = seq_df_sort[seq_df_sort['best_test_iou'] > ]
print(seq_df_sort[['step', 'best_test_iou']])
print(seq_df_sort)

for name in namelist:
    colname = f'best_test_iou{name}'
    plt.figure(figsize=(7,5))
    plt.title(colname)
    plt.ylabel('IoU')
    plt.xlabel('run')
    ymin = seq_df_sort.loc[:, colname].min()
    ymax = seq_df_sort.loc[:, colname].max()
    y0 = ymin - (ymax-ymin)*0.5
    print(y0)
    plotted_models = []
    for idx in seq_df_sort.index[:]:
        # print(name, idx, seq_df_sort.loc[idx, colname].min(), seq_df_sort.loc[idx, colname].max())
        # y0 = seq_df_sort.loc[idx, colname].min()
        barheight = seq_df_sort.loc[idx, colname]
        y = barheight - y0
        x = idx+1
        # print(y)
        modelname = seq_df_sort.loc[idx, 'model']
        # print(seq_df_sort.loc[idx])
        if modelname in plotted_models:
            # plt.bar(idx, y, width=0.8, bottom=y0, color=f'C{modelname[0]}-1')
            label = ''
        else:
            # plt.bar(idx, y, width=0.8, bottom=y0, color=f'C{modelname[0]}-1', label=f'{modelname}')
            label=f'{modelname}'
        plt.bar(x, y, width=0.8, bottom=y0, color=f'C{int(modelname[0])-1}', label=label)

        # plt.ylim(top=barheight*1.015)
        plottext = f'lrn: {seq_df_sort.loc[idx, "learning_rate"]:.0e}'
        # plottext = f'step: {seq_df_sort.loc[idx, "step"]}, lrn: {seq_df_sort.loc[idx, "learning_rate"]:.0e}'
        # if modelname == '1_model_unet':
        #     plottext += f', layers: {seq_df_sort.loc[idx, "num_layers"]}, feat: {seq_df_sort.loc[idx, "coef_num_features"]}'

        # plt.text(x-0.2, y0+(ymax-ymin)*0.05, plottext, rotation='vertical')
        # plt.bar(idx, y, width=0.5, bottom=y0)
        plotted_models.append(modelname)
    bottom, top = plt.ylim()
    print(bottom, top)
    plt.ylim(top=top + (top-bottom)*0.25)
    plt.xticks(seq_df_sort.index[:]+1)
    plt.legend(loc='best', ncol=2)
    plt.tight_layout()
    plt.savefig(f'iou_plots/{colname}.png')

# plt.bar(cols, seq_df.loc[0,cols], width=0.5, align='edge', label='unet 3 3')
# plt.bar(cols, seq_df.loc[1,cols], width=0.5, label='unet 3 4')
# plt.ylabel('IoU')
# plt.figure(figsize=(7,5))
# for idx in seq_df.index:
#     plt.figure(figsize=(7, 5))
#     plt.bar(cols, seq_df.loc[idx,cols], width=0.5, label=idx)
# plt.ylabel('IoU')

plot_idx = 0
# best_run_name = seq_df_sort.loc[plot_idx, 'run_name']
# best_model_name = seq_df_sort.loc[plot_idx, 'model']
# best_learning_rate = seq_df_sort.loc[plot_idx, 'learning_rate']


plt.figure(figsize=(7, 5))
plt.bar([col[10:] for col in cols], seq_df.loc[plot_idx, cols], width=0.5)
plt.ylabel('IoU')
plt.title(f'{seq_df_sort.loc[plot_idx, "model"]}, learning_rate={seq_df_sort.loc[plot_idx, "learning_rate"]}')
plt.tight_layout()
plt.savefig(f'iou_plots/iou_{sequence_name}_{plot_idx}.png')

# seq_df[cols].plot()
# plt.legend()
plt.show()

