import random

import matplotlib.pyplot as plt
import torch
import torchvision
import os
import numpy as np
import json
import torchvision.transforms as T
import torch.nn.functional as F
import torchvision.transforms.functional as TF


class AddGaussianNoise(object):
    def __init__(self, mean=0., std=1.):
        self.std = std
        self.mean = mean

    def __call__(self, tensor):
        return tensor + torch.randn(tensor.size()) * self.std + self.mean

    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1})'.format(self.mean, self.std)

class AddRandomNoise(object):
    def __init__(self, mean=(0., 1.), std=(0., 1.)):
        self.std = np.random.uniform(std[0], std[1])
        self.mean = np.random.uniform(mean[0], mean[1])
        print(self.mean, self.std)

    def __call__(self, tensor):
        return tensor + torch.randn(tensor.size()) * self.std + self.mean

    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1})'.format(self.mean, self.std)


mmap_dir = './celeb_hq_dataset'
# mmap_dir = './celeb_hq_dataset-test'
mmap_path = os.path.join(mmap_dir, 'data.mmap')

json_path = os.path.join(mmap_dir, 'data.json')
with open(json_path) as fp:
    parameters = json.load(fp)

# print(f'parameters: {parameters}')

shape = tuple(parameters['shape'])
channels = parameters['channels']

mmap = np.memmap(mmap_path, dtype='float16', mode='r', shape=shape)


channel_grayscale = channels.index('photo_grayscale')
channel_mask1 = channels.index('eye')

angle = random.random()*360
print(angle)

idx = 0
x = torch.FloatTensor(np.array(mmap[idx, channel_grayscale:channel_grayscale+1, :, :])) # face
y = torch.FloatTensor(np.array(mmap[idx, channel_mask1:, :, :]))  # mask
# y_other = torch.ones([1, y.size(1), y.size(2)]) - torch.sum(y, dim=0)
# y_other[y_other < 1] = 0
# y = torch.cat((y, y_other), dim=0)
# x = torch.FloatTensor(np.array(mmap[idx, channel_grayscale+1:channel_grayscale+2, :, :])) # face
# x += abs(min(x))
xy_cat = torch.cat((x, y), dim=0)
# print(xy_cat, type(xy_cat))
orig_size = x.shape[-1]
print(orig_size)

# F.rotate(img, angle, self.resample, self.expand, self.center, self.fill)

# plt.figure()
# plt.imshow(x[0])
trans = T.Compose([
    # T.RandomResizedCrop(size=(64, 64)),
    # T.RandomRotation(degrees=90),
    # T.RandomHorizontalFlip(p=0.5),
    T.RandomCrop(size=(200, 200)),
    T.RandomAffine(degrees=(-90, 90), translate=(0.1, 0.3), scale=(0.8, 1.2))
])

transforms = torch.nn.ModuleList([
    # T.RandomResizedCrop(size=(200, 200)),
    # T.RandomRotation(degrees=90),
    T.RandomHorizontalFlip(p=0.5),
    T.RandomCrop(size=(200, 200)),
    T.RandomAffine(degrees=(-70, 70), translate=(0.1, 0.3), scale=(0.8, 1.2)),
    # AddGaussianNoise(0., 1.)
])
trans = torchvision.transforms.RandomApply(
    transforms,
    p=0.9)

add_noise = torchvision.transforms.RandomApply(
    [AddRandomNoise(mean = (0., 0.25), std=(0., 1.))],
    p=0.9)



# x = x.transpose()
# xy_cat = trans(xy_cat)
#
# x = xy_cat[:1,:,:]
# y = xy_cat[1:,:,:]

mean_range = (0., 0.)
std_range = (0., 1.)

plt.figure()
row_number = 5
for rand_idx in range(row_number):
    xy_cat_rot = trans(xy_cat)
    # std = torch.randn(1) * (std_range[1] - std_range[0]) + std_range[0]
    # mean = torch.randn(1) * (mean_range[1] - mean_range[0]) + mean_range[0]
    # print(mean, std)
    add_noise = torchvision.transforms.RandomApply(
        [AddRandomNoise(mean=(0., 0.), std=(0., 0.5))],
        p=0.9)
    x = add_noise(xy_cat_rot[:1,:,:])
    print(x.size())
    down_size = np.random.randint(int(orig_size/8), int(orig_size/2))
    resample = torchvision.transforms.RandomApply(
        torch.nn.ModuleList([
            T.Resize(size=down_size),
            T.Resize(size=orig_size)
        ]),
        p=0.9)
    x = resample(x)
    print(x.size())
    y = xy_cat_rot[1:,:,:]
    y_other = torch.ones([1, y.size(1), y.size(2)]) - torch.sum(y, dim=0)
    y_other[y_other < 1] = 0
    y = torch.cat((y, y_other), dim=0)
    dim = len(x) + len(y)
    # print(dim)
    plt.subplot(row_number, dim, rand_idx*dim + 1)
    plt.imshow(x[0], cmap=plt.get_cmap('gray'))
    for idx, pic in enumerate(y):
        plt.subplot(row_number, dim, rand_idx*dim + idx+2)
        plt.imshow(pic)
        # plt.imshow(x[0, :, :], cmap=plt.get_cmap('gray'))


plt.show()